#!/usr/bin/perl
use DB_File;
use Fcntl;
use strict;

#locking routine from camel book
sub LOCK_SH { 1 }
sub LOCK_EX { 2 }
sub LOCK_NB { 4 }
sub LOCK_UN { 8 }

sub SECS_WK { 604800 }
sub SECS_DAY { 86400 }
sub SECS_HR { 3600 }
sub SECS_MIN { 60 }


my %MP3s;
my $bdname="$ENV{'HOME'}/.mp3stats";
my $maxlines=-1;
my $db = tie %MP3s, "DB_File", $bdname, O_RDWR|O_CREAT, 0644, $DB_HASH or die "Can't open database $bdname";
my @summaries;
my $fd=$db->fd;
my $quiet=1;
print STDERR "$$: Locking $bdname\n" unless $quiet;
open(DB_FH, "+<&=$fd") or die "fdopen $!";

unless (flock (DB_FH, LOCK_EX | LOCK_NB)) {
    print STDERR "$$: CONTENTION; waiting for lock ($!) ..." unless $quiet;
    unless (flock (DB_FH, LOCK_EX)) { die "flock: $!" }
}
print STDERR "$$: Lock granted\n" unless $quiet;


my $now = time();
# by default a summary queries the last 10 days.
my ($since,$before) = ($now-7*SECS_DAY,$now);

my $i;
for ($i=0; $i<=$#ARGV;) {
    if ($ARGV[$i] =~ /--summary/) {
	doSummary();
	splice(@ARGV, $i, 1);
    } elsif ($ARGV[$i] =~ /--push/) {
	pushFile($ARGV[$i+1]);
	splice(@ARGV, $i, 2);
    } elsif ($ARGV[$i] =~ /--maxlines/) {
	$maxlines=$ARGV[$i+1];
	splice(@ARGV, $i, 2);
    } elsif ($ARGV[$i] =~ /--since/) {
	$since=adjusttime($ARGV[$i+1]);
	splice(@ARGV, $i, 2);
    } elsif ($ARGV[$i] =~ /--before/) {
	$before=adjusttime($ARGV[$i+1]);
	splice(@ARGV, $i, 2);
    } elsif ($ARGV[$i] =~ /--compare/) {
	compareSummaries();
	splice(@ARGV, $i, 1);
    } elsif ($ARGV[$i] =~ /--print/) {
	printSummary();
	splice(@ARGV, $i, 1);
    } else {
	$i++;
    }
}
print STDERR "Unprocessed args: ", join ' ', @ARGV, "\n" if ($#ARGV>=0);

$db->sync();
flock(DB_FH, LOCK_UN);
untie %MP3s;
undef $db;

sub pushFile() {
    my $name=shift;
    if (defined $MP3s{$name}) {
	$MP3s{$name}.=':'.time();
    } else {
	$MP3s{$name}='0:'.time();
    }
    print STDERR "Pushed: $name: ",$MP3s{$name}, "\n" unless $quiet;
}


sub printSummary() {
    my $s;
    my $lines=0;

    my $summary=pop @summaries;
    foreach $s (@{$summary}) {
	print "$s\n" if ($lines<$maxlines || $maxlines<0);
	$lines++;
    }
    $db->sync();
    flock(DB_FH, LOCK_UN);
    exit;
}

sub compareSummaries() {
    my @newsummary;
    my $thistime=pop @summaries;
    my $lasttime=pop @summaries;

    my ($line1, $line2);
    foreach $line1 (@{$thistime}) {
	my $posn=1;
	my $found=0;
	foreach $line2 (@{$lasttime}) {
	    if ($line1 eq $line2) {
		$line1.="\t$posn";
		$found=1;
	    }
	    $posn++;
	}
#	print STDERR "Found: $found\n";
	$line1.="\t--" if (!$found);
	push @newsummary, $line1;
    }
    push @summaries, [ @newsummary ];
}

sub adjusttime() {
    my ($adjustment)=@_;
    $adjustment =~ /(.*)w/ and $adjustment=$now-$1*SECS_WK;
    $adjustment =~ /(.*)d/ and $adjustment=$now-$1*SECS_DAY;
    $adjustment =~ /(.*)m/ and $adjustment=$now-$1*SECS_MIN;
    $adjustment =~ /(.*)h/ and $adjustment=$now-$1*SECS_HR;
    $adjustment =~ /(.*)s/ and $adjustment=$now-$1;
    return $adjustment;
}


sub doSummary() {
    my @summary;
    my $key;
    print STDERR "Now: ", scalar localtime($now),
	" Since: ", scalar localtime($since),
	    " Before: ", scalar localtime($before), "\n" unless $quiet;
    foreach $key (keys %MP3s) {
	my @times=split /:/, $MP3s{$key};
	my $t;
	$times[0]=0;
	for ($t=1; $t<=$#times; $t++) {
	    splice @times, $t, 1 if ($times[$t]-$times[$t-1]<5);
	    next if ($t>$#times);
	    $times[0]++ if ($times[$t]>$since && $times[$t]<$before);
	}
	$MP3s{$key}=join ':', @times;
#	print "$key total: $times[0]\n";
    }
#    print "Sorting..\n";
    foreach $key (reverse sort totalcmp keys %MP3s) {
	my @times=split /:/, $MP3s{$key};
	next if $key =~/null/;
	next if $key =~/CD Audio/;
	push @summary, $key if $times[0];
#	print "$key\n" if $times[0];
#	foreach $time (@times) {
#	    print "\t", scalar localtime($time), "\n" if ($time > $since);
#	}
    }
    push @summaries, [@summary];
}

sub totalcmp { 
    my @ta = split /:/, $MP3s{$a};
    my @tb = split /:/, $MP3s{$b};
# comparing the last elems gives the more recent the preference
    $ta[0] <=> $tb[0] or $ta[$#ta] <=> $tb[$#tb];	    
}

